from __future__ import print_function

from ghizmo.commands import lib

def stale_pr_branches(config, args):
  """
  List "stale" PR branches, i.e. those for a closed PR from the same, non-forked repository.
  """
  repo = config.repo
  for pr in repo.pull_requests(state="closed"):
    if pr.head.repo == pr.base.repo and repo.branch(pr.head.ref):
      yield {
        "html_url": pr.html_url,
        "base_branch": pr.base.ref,
        "head_branch": pr.head.ref,
      }



"""
Ghizmo main library.
"""

from __future__ import print_function

__author__ = 'jlevy'

import logging as log
import sys
import json
import yaml
import inspect
import os
import importlib
import getpass
from collections import namedtuple
from collections import OrderedDict
from functools32 import lru_cache  # functools32 pip
import github3  # github3.py pip
from github3.null import NullObject

import configs
import commands


def read_login_info(username=None):
  if not username:
    username = raw_input("GitHub username: ")
  return (username, getpass.getpass())


def login(username=None):
  token = configs.get_access_token()
  if token:
    log.info("Using access token authentication")
    return github3.login(token=token)
  else:
    username = username or configs.get_username()
    return github3.login(*read_login_info(username=username))


def format_to_string(obj, format=None):
  if not format:
    format = "json"

  if isinstance(obj, NullObject):
    raise ValueError("Command returned null type (invalid input?): %s" % repr(obj))

  # Heuristic to convert github3 objects to serializable form.
  if hasattr(obj, "as_dict"):
    obj = obj.as_dict()

  if format == "json":
    return json.dumps(obj, sort_keys=True, indent=2) + "\n"
  elif format == "yaml":
    return yaml.safe_dump(obj, default_style='"')
  else:
    raise AssertionError("Invalid format: %s" % format)


def print_formatter(format=None):
  def printer(obj):
    print(format_to_string(obj, format), end="")
    sys.stdout.flush()

  return printer

# The configuration to run.
Config = namedtuple("Config", "repo formatter")


def _to_dash(name):
  return name.replace("_", "-")


def _to_underscore(name):
  return name.replace("-", "_")


def _is_public_func(f):
  return inspect.isfunction(f) and not f.__name__.startswith("_")


@lru_cache()
def all_command_functions():
  # TODO: Clean this up somehow.
  # Use absolute module path so it works post installation.
  modules = [importlib.import_module("ghizmo.commands.%s" % module) for module in commands.__all__]

  # If there is a ghizmo_commands.py file in the current directory, use it too.
  if os.path.exists("ghizmo_commands.py"):
    modules.append(importlib.import_module("ghizmo_commands"))

  log.info("Imported command modules: %s", modules)

  func_map = {}
  for module in modules:
    for (name, func) in inspect.getmembers(module, predicate=_is_public_func):
      if name in func_map:
        raise ValueError("Duplicate function name for command: %s" % name)
      func_map[name] = func
  # Sort items first by module, then by name.
  return OrderedDict(sorted(func_map.items(), key=lambda (name, func): (func.__module__, name)))


@lru_cache()
def command_directory(use_dashes=True):
  def doc_for_function(func):
    if func.__doc__:
      return func.__doc__.strip()
    else:
      return "(no pydoc)"

  transform = _to_dash if use_dashes else lambda x: x
  return [(func.__module__, transform(name), doc_for_function(func)) for (name, func) in
          all_command_functions().iteritems()]


@lru_cache()
def list_commands(use_dashes=True):
  return [command for (module, command, doc) in command_directory(use_dashes=use_dashes)]


def get_command_func(command):
  command = _to_underscore(command)
  if not command in all_command_functions():
    raise ValueError("invalid command: %s" % command)
  return all_command_functions()[command]


def run_command(command, config, args):
  command_func = get_command_func(command)
  log.info("Command '%s' (%s)", command, command_func)
  log.info("Config: %s", config)
  log.info("Args: %s", args)
  # This executes the command step by step, either just to display results, or to display progress
  # on an action with side effects.
  for result in command_func(config, args):
    config.formatter(result)


# TODO:
# Proper streaming of JSON items (as opposed to line-by-line), for use in piping.
# Prettier SIGPIPE handling



from __future__ import print_function

from ghizmo.commands import lib


def tags(config, args):
  """
  List all tags.
  """
  return config.repo.tags()


def show_tags(config, args):
  """
  Show info for tags supplied on stdin.
  """
  for item in lib.input_json_lines():
    yield config.repo.tag(item)


def _delete_ref(repo, ref_name, force, dry_run):
  ref = repo.ref(ref_name)
  if not ref and not force:
    raise ValueError("Reference not found: %s" % ref_name)
  if not dry_run:
    ref.delete()
  return lib.status("Deleted %s" % ref_name, dry_run=dry_run)


def branches(config, args):
  """
  List all branches.
  """
  return config.repo.branches()


def show_branches(config, args):
  """
  Show branches supplied on stdin.
  """
  for item in lib.input_json_lines():
    yield config.repo.branch(item)


def delete_branches(config, args):
  """
  Delete branches supplied on stdin.
  """
  for ref_name in lib.input_json_lines():
    yield _delete_ref(config.repo, "heads/" + ref_name, args.force, args.dry_run)


def refs(config, args):
  """
  List all refs.
  """
  return config.repo.refs()


def show_refs(config, args):
  """
  Show refs supplied on stdin.
  """
  for item in lib.input_json_lines():
    yield config.repo.ref(item)


def delete_refs(config, args):
  """
  Delete refs supplied on stdin.
  """
  for ref_name in lib.input_json_lines():
    yield _delete_ref(config.repo, ref_name, args.force, args.dry_run)


def pull_requests(config, args):
  """
  List all PRs.
  """
  state = args.state or "open"
  return config.repo.pull_requests(state=state)


def contributors(config, args):
  """
  List all contributors.
  """
  return config.repo.contributors()


def collaborators(config, args):
  """
  List all collaborators.
  """
  return config.repo.collaborators()



"""
Common functions for commands.
"""

from __future__ import print_function

__author__ = 'jlevy'

# TODO: Move this module elsewhere?

import sys
import json


def status(message, dry_run=False):
  return {"message": message, "dry_run": dry_run}


def process_input(function):
  """
  Just for brevity, a generic handler to look up and process based on input.
  """
  for item in input_json_lines():
    yield function(item)


def input_json_lines():
  """
  Read line-delimited JSON stream.
  TODO: Handle jq-style streamed JSON that need not be line delimited.
  """
  for line in sys.stdin:
    yield json.loads(line)



"""
This module holds the pre-defined commands.

Public functions here (not starting with '_') are auto-registered into Ghizmo's command line.

Best practices:
- Include a one-line pydoc, which will appear in the --help output.
- Yield results for items that search for information.
- Yield status for items that do things.

For the GitHub library API, see: https://github.com/sigmavirus24/github3.py
"""

__author__ = 'jlevy'

__all__ = ['core', 'general']





"""
Ghizmo config handling.
"""

from __future__ import print_function

__author__ = 'jlevy'

import logging as log
import os
import re
import yaml
import subprocess

from functools32 import lru_cache  # functools32 pip

CONFIG_FILENAME = ".ghizmo.yml"


@lru_cache()
def load_config_file():
  path = os.path.join(os.environ["HOME"], CONFIG_FILENAME)

  parsed_configs = None
  if os.path.exists(path):
    with open(path) as f:
      parsed_configs = yaml.safe_load(f)

  return parsed_configs


def get_username():
  configs = load_config_file()
  return configs and configs["username"]


def get_access_token():
  """Return GitHub auth token, if in config file."""
  configs = load_config_file()
  return configs and configs["access_token"]


def _extract_github_repo_info(url):
  m = re.match("^git@github.com:([a-zA-Z0-9_-]+)/([a-zA-Z0-9_-]+).git$", url) or \
      re.match("^https://github.com/([a-zA-Z0-9_-]+)/([a-zA-Z0-9_-]+).git$", url)
  return (m.group(1), m.group(2)) if m else None


def infer_repo(remote_name="origin"):
  """
  Extract the current repository info, if available, using .git/config in current working directory.
  """
  log.info("Checking in cwd for git dirctory: %s", os.getcwd())
  remote_url = subprocess.check_output(["git", "config", "--get", "remote.%s.url" % remote_name])
  (owner, repo_name) = _extract_github_repo_info(remote_url)
  log.info("Inferred repository: %s/%s", owner, repo_name)
  return (owner, repo_name)



#!/usr/bin/env python
"""
If commands require input, it must be line-delimited JSON (e.g. quoted strings).

For further documentation, see: https://github.com/jlevy/ghizmo
"""

from __future__ import print_function

__author__ = 'jlevy'

import logging as log
import sys
import argparse

NAME = "ghizmo"
VERSION = "0.1.3"
DESCRIPTION = "ghizmo: Extensible GitHub command-line tricks"
LONG_DESCRIPTION = __doc__


def log_setup(level):
  if level == log.DEBUG:
    log.basicConfig(format="%(levelname).1s %(filename)20s:%(lineno)-4d  %(message)s", level=level,
                    stream=sys.stderr)
  else:
    log.basicConfig(format="%(message)s", level=level, stream=sys.stderr)

    def brief_excepthook(exctype, value, traceback):
      print("error: %s" % value, file=sys.stderr)
      print("(run with --debug for traceback info)", file=sys.stderr)
      sys.exit(2)

    sys.excepthook = brief_excepthook


def main():
  import ghizmo
  import configs

  command_directory = ghizmo.command_directory()
  command_modules = sorted(set([module for (module, name, doc) in command_directory]))
  command_docs = \
    "command modules: %s\n" % ", ".join(command_modules) \
    + "(augment by adding to ./ghizmo_commands.py)\n\n" \
    + "commands available:\n" \
    + "\n".join(["  %s: %s" % (name, doc) for (module, name, doc) in command_directory])
  parser = argparse.ArgumentParser(description=DESCRIPTION, version=VERSION,
                                   epilog="\n" + __doc__ + "\n" + command_docs,
                                   formatter_class=argparse.RawTextHelpFormatter)
  parser.add_argument("command", help="%s command" % NAME, choices=ghizmo.list_commands())
  parser.add_argument("--state", help="pull requests state", choices=["open", "closed", "all"])
  parser.add_argument("--format", help="output format", choices=["json", "yaml"])
  parser.add_argument("--username", help="username to log in as")
  parser.add_argument("--repo", help="repo of the form: owner/repo-name")
  parser.add_argument("-f", "--force", help="enable debugging output", action="store_true")
  parser.add_argument("-n", "--dry-run", help="dry run: log actions but don't do anything", action="store_true")
  parser.add_argument("--debug", help="enable debugging output", action="store_true")
  args = parser.parse_args()

  log_setup(log.DEBUG if args.debug else log.WARN)

  # Validate credentials and log in.
  gh = ghizmo.login(username=args.username)
  if not gh:
    raise ValueError("Login failure")

  # Validate repository.
  if args.repo:
    (owner, repo_name) = args.repo.split("/")
  else:
    (owner, repo_name) = configs.infer_repo()
  repo = gh.repository(owner, repo_name)
  if not repo:
    raise ValueError("Couldn't access repository: %s/%s" % (owner, repo_name))

  # Assemble config for this run.
  formatter = ghizmo.print_formatter(args.format)
  config = ghizmo.Config(repo=repo, formatter=formatter)

  ghizmo.run_command(args.command, config, args)


if __name__ == '__main__':
  main()



#!/usr/bin/env

from setuptools import setup, find_packages
from ghizmo import main

setup(
  name="ghizmo",
  version=main.VERSION,
  packages=find_packages(),
  author="Joshua Levy",
  license="Apache 2",
  url="https://github.com/jlevy/ghizmo",
  install_requires=["github3.py>=1.0.0a2", "PyYAML>=3.11", "functools32>=3.2.3"],
  description=main.DESCRIPTION,
  long_description=main.LONG_DESCRIPTION,
  classifiers=[
    'Development Status :: 4 - Beta',
    'Environment :: Console',
    'Intended Audience :: End Users/Desktop',
    'Intended Audience :: System Administrators',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: Apache Software License',
    'Operating System :: MacOS :: MacOS X',
    'Operating System :: POSIX',
    'Operating System :: Unix',
    'Programming Language :: Python :: 2.7',
    'Topic :: Utilities',
    'Topic :: Software Development'
  ],
  entry_points={
    "console_scripts": [
      "ghizmo = ghizmo.main:main",
    ],
  },
)
